package de.uniulm.mathematik.typo.afm;

import java.io.*;
import java.util.*;

/**
 * Instances of this class represent individual Adobe Font Metrics.
 */
public class AdobeFontMetrics implements FontMetrics {
   private Parser parser;
   private PropertyTable properties;
   private CharacterMetricsTable cms;
   private int minCode, maxCode;
   private CharacterMetrics[] cm;
   private KerningPairTable kpairs;

   /**
    * Construct a font metrics from an Adobe Font Metrics source (AFM).
    * @param in a buffered reader that is connected to a source file.
    */
   public AdobeFontMetrics(BufferedReader in) throws Parser.ParserException {
      Lex lex = new Lex(in);
      parser = new Parser(lex);
      properties = parser.getProperties();
      cms = parser.getCharacterMetrics();
      cm = new CharacterMetrics[cms.maxCode() + 1];
      minCode = cms.minCode(); maxCode = cms.maxCode();
      for (int code = minCode; code <= maxCode; ++code) {
	 cm[code] = cms.get(code);
      }
      kpairs = parser.getKerningPairs();
   }

   /**
    * Construct a font metrics from an Adobe Font Metrics source (AFM).
    * @param filename path name of an AFM file.
    */
   public AdobeFontMetrics(String filename)
	 throws Parser.ParserException, FileNotFoundException {
      this(new BufferedReader(new FileReader(filename)));
   }

   private String get(String property) {
      try {
	 String value = (String) properties.get(property);
	 return value;
      } catch (PropertyTable.PropertyTableException e) {
	 return null;
      }
   }

   private int getInt(String property) {
      String value = get(property);
      if (value == null) {
	 return 0;
      } else {
	 try {
	    Scanner s = new Scanner(value);
	    return s.nextInt();
	 } catch (Exception e) {
	    return 0;
	 }
      }
   }

   /**
    * Return the name of the font, e.g. "Times-Roman".
    */
   public String getName() {
      return get("FontName"); 
   }

   /**
    * Return the name of a character.
    */
   public String getName(int code) {
      if (defined(code)) {
	 return cm[code].getName();
      } else {
	 return null;
      }
   }

   /**
    * Return lowest code number.
    */
   public int minCode() {
      return minCode;
   }

   /**
    * Return highest code number.
    */
   public int maxCode() {
      return maxCode;
   }

   /**
    * Return number of characters within this font.
    */
   public int length() {
      return cms.length();
   }

   /**
    * Return true, if there is a character with the given code.
    */
   public boolean defined(int code) {
      return code >= minCode && code <= maxCode && cm[code] != null;
   }

   /**
    * Return font bounding box as array of four integers:
    * llx, lly, urx, and ury.
    */
   public int[] getBoundingBox() {
      int[] bbox = new int[4];
      String fontbbox = get("FontBBox");
      if (fontbbox != null) {
	 Scanner s = new Scanner(fontbbox);
	 for (int i = 0; i < 4; ++i) {
	    bbox[i] = s.nextInt();
	 }
      }
      return bbox;
   }

   /**
    * Return bounding box of an individual character as array of
    * four integers: llx, lly, urx, and ury. Zeroes are returned
    * if there is no character with the given code or no bbox
    * has been provided for this character.
    */
   public int[] getBoundingBox(int code) {
      if (defined(code)) {
	 return cm[code].getBoundingBox();
      } else {
	 return new int[4];
      }
   }

   /**
    * Return the versal height of the font, e.g. the height of
    * versal letters.
    */
   public int getVersalHeight() {
      return getInt("CapHeight");
   }

   /**
    * Return the height of the lower case letter x.
    */
   public int getXHeight() {
      return getInt("XHeight");
   }

   /**
    * Return the height between the baseline and the bottom of the
    * lower case letter p.
    */
   public int getDescender() {
      return getInt("Descender");
   }

   /**
    * Return the height between the baseline and the height of the
    * lower case letter d.
    */
   public int getAscender() {
      return getInt("Ascender");
   }

   /**
    * Return width of an individual character. Zero is returned
    * if there is no character with the given code or no width
    * has been specified for this character.
    */
   public int getWidth(int code) {
      if (defined(code)) {
	 return cm[code].getWX();
      } else {
	 return 0;
      }
   }

   /**
    * Return the kerning for the given two characters. 0 is
    * returned if no kerning has been defined.
    */
   public int getKerning(int code1, int code2) {
      if (kpairs.defined(code1, code2)) {
	 KerningPair pair = kpairs.get(code1, code2);
	 return pair.getKerningX();
      } else {
	 return 0;
      }
   }
}
