package de.uniulm.mathematik.typo.afm;

import java.lang.*;

/**
 * An instance of this class represents an individual kerning pair.
 * See also: Adobe Font Metrics File Specification, section 9.2.
 */
public class KerningPair {
   private int code1;
   private int code2;
   private String name1;
   private String name2;
   private int offsetx;
   private int offsety;
   private String line;
   private Lex lex;

   /**
    * Inner class for exceptions raised by the KerningPair class.
    */
   public static class ParserException extends Exception {
      /**
       * The source line which triggered the syntax error.
       */
      public final String line;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * token and a printable message within a throw statement.
       * @param line corresponding AFM source line.
       * @param message printable error message.
       */
      public ParserException (String line, String message) {
	 this.line = line;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the source line and the error message.
       * @return the string with the error message.
       */
      public String toString() {
	 return message + ": " + line;
      }
   } // class ParserException

   private class Lex {
      private String line;
      private int index;
      private char ch;
      private boolean eof;
      public final static int EOF = 0;
      public final static int ERROR = 1;
      public final static int IDENTIFIER = 2;
      public final static int INTEGER = 3;
      public int sy;
      public String ident;
      public int intval;

      private void nextch() {
	 if (!eof) {
	    if (index == line.length()) {
	       eof = true;
	    } else {
	       ch = line.charAt(index);
	       ++index;
	    }
	 }
      }

      public void getsy() {
	 while (!eof && Character.isWhitespace(ch)) {
	    nextch();
	 }
	 if (eof) {
	    sy = EOF;
	 } else if (Character.isLetter(ch)) {
	    ident = (new Character(ch)).toString();
	    nextch();
	    while (!eof && !Character.isWhitespace(ch)) {
	       ident += ch;
	       nextch();
	    }
	    sy = IDENTIFIER; 
	 } else if (ch == '-' || Character.isDigit(ch)) {
	    String value = (new Character(ch)).toString();
	    nextch();
	    while (!eof && Character.isDigit(ch)) {
	       value += ch;
	       nextch();
	    }
	    try {
	       intval = java.lang.Integer.parseInt(value, 10);
	       sy = INTEGER;
	    } catch (NumberFormatException e) {
	       sy = ERROR;
	    }
	 } else if (ch == '<') {
	    String value = "";
	    nextch();
	    while (!eof && ch != '>') {
	       value += ch;
	       nextch();
	    }
	    try {
	       intval = java.lang.Integer.parseInt(value, 16);
	       sy = INTEGER;
	    } catch (NumberFormatException e) {
	       sy = ERROR;
	    }
	    nextch();
	 } else {
	    sy = ERROR;
	 }
      }

      public Lex(String line) {
	 this.line = line;
	 index = 0;
	 eof = false;
	 nextch();
      }
   }

   private void expect(int symbol, String what) throws ParserException {
      if (lex.sy != symbol) {
	 throw new ParserException(line, what + " expected");
      }
   }

   private int fetchint(String what) throws ParserException {
      expect(Lex.INTEGER, "integer");
      int value = lex.intval;
      lex.getsy();
      return value;
   }

   private String fetchident(String what) throws ParserException {
      expect(Lex.IDENTIFIER, what);
      String ident = lex.ident;
      lex.getsy();
      return ident;
   }

   private void parseKerningPair() throws ParserException {
      String keyword = fetchident("keyword");
      if (keyword.equals("KP")) {
	 name1 = fetchident("name1");
	 name2 = fetchident("name2");
	 offsetx = fetchint("number1");
	 offsety = fetchint("number2");
      } else if (keyword.equals("KPH")) {
	 code1 = fetchint("hex1");
	 code2 = fetchint("hex2");
	 offsetx = fetchint("number1");
	 offsety = fetchint("number2");
      } else if (keyword.equals("KPX")) {
	 name1 = fetchident("name1");
	 name2 = fetchident("name2");
	 offsetx = fetchint("number");
	 offsety = 0;
      } else if (keyword.equals("KPY")) {
	 name1 = fetchident("name1");
	 name2 = fetchident("name2");
	 offsetx = 0;
	 offsety = fetchint("number");
      } else {
	 throw new ParserException(keyword, "invalid keyword for kerning pair");
      }
   }

   private void parse() throws ParserException {
      parseKerningPair();
      if (lex.sy != Lex.EOF) {
	 throw new ParserException(line, "syntax error");
      }
   }

   /**
    * Return the internal data structures as printable string.
    */
   public String toString() {
      if (name1 != null) {
	 return "kerning pair (" + name1 + ", " + name2 + ") = (" +
	    offsetx + ", " + offsety + ")";
      } else {
	 return "kerning pair (<" + java.lang.Integer.toString(code1, 16) +
	    ">, <" + java.lang.Integer.toString(code2) + ">) = (" +
	    offsetx + ", " + offsety + ")";
      }
   }

   /**
    * Return the name of the first glyph of the kerning pair.
    * Null is returned if the pair was specified by encodings
    * instead of names.
    */
   public String getName1() {
      return name1;
   }

   /**
    * Return the name of the second glyph of the kerning pair.
    * Null is returned if the pair was specified by encodings
    * instead of names.
    */
   public String getName2() {
      return name2;
   }

   /**
    * Return the encoding value of the first glyph of the kerning pair.
    * -1 is returned if the pair was specified by names instead
    * of encoding values.
    */
   public int getCode1() {
      return code1;
   }

   /**
    * Return the encoding value of the second glyph of the kerning pair.
    * -1 is returned if the pair was specified by names instead
    * of encoding values.
    */
   public int getCode2() {
      return code2;
   }

   /**
    * Return the x coordinate of the kerning vector.
    */
   public int getKerningX() {
      return offsetx;
   }

   /**
    * Return the y coordinate of the kerning vector.
    */
   public int getKerningY() {
      return offsety;
   }

   /**
    * Construct a kerning pair from the given AFM source line.
    * @param line the AFM source line.
    */
   public KerningPair(String line) throws ParserException {
      this.line = line;
      name1 = null; name2 = null;
      code1 = -1; code2 = -1;
      lex = new Lex(line);
      lex.getsy();
      parse();
   }
} // class KerningPair
