package de.uniulm.mathematik.typo.afm;

import java.util.*;

/**
 * Instances of this class represent tables of kerning pairs.
 * The AFM source file specifies kerning pairs either as pairs
 * of encoding values or as pairs of names. Firstly, this table
 * stores the pairs under the keys as given. If the resolve
 * method is used, pairs can be looked up either by encoding
 * values or name pairs, if the corresponding character metrics
 * entries support both.
 * See Adobe Font Metrics File Specification, section 9.2.
 */
public class KerningPairTable {
   private HashMap<Integer, KerningPair> byCode;
   private HashMap<String, KerningPair> byName;
   private LinkedList<KerningPair> resolveName;
   private LinkedList<KerningPair> resolveCode;

   /**
    * Inner class for exceptions raised by the CharacterMetricsTable class.
    */
   public static class KerningPairTableException extends Exception {
      /**
       * The key.
       */
      public final String key;

      /**
       * A printable error message.
       */
      public final String message;

      /**
       * Construct an exception object encapsulating a
       * token and a printable message within a throw statement.
       * @param key key string.
       * @param message printable error message.
       */
      public KerningPairTableException(String key, String message) {
	 this.key = key;
	 this.message = message;
      }

      /**
       * Return a string containing an error message which
       * includes both, the key and the error message.
       * @return the string with the error message.
       */
      public String toString() {
	 if (key != null) {
	    return message + ": " + key;
	 } else {
	    return message;
	 }
      }
   } // class KerningPairTableException

   /**
    * Construct a table of kerning pairs that is initially empty.
    */
   public KerningPairTable() {
      byCode = new HashMap<Integer, KerningPair>();
      byName = new HashMap<String, KerningPair>();
      resolveName = new LinkedList<KerningPair>();
      resolveCode = new LinkedList<KerningPair>();
   }

   private String makeKey(String name1, String name2) {
      return name1 + " " + name2;
   }

   private Integer makeKey(int code1, int code2) {
      return new Integer(code1 * 256 + code2);
   }

   /**
    * Add the given kerning pair to the table. Note that
    * the pair must not have yet been inserted into the table.
    * @param pair the to be inserted kerning pair.
    */
   public void add(KerningPair pair)
	 throws KerningPairTableException {
      String name1 = pair.getName1();
      String name2 = pair.getName2();
      if (name1 != null) {
	 String key = makeKey(name1, name2);
	 if (byName.containsKey(key)) {
	    throw new KerningPairTableException(key,
	       "multiply defined kerning pair");
	 }
	 byName.put(key, pair);
      } else {
	 resolveName.add(pair);
      }
      int code1 = pair.getCode1();
      int code2 = pair.getCode2();
      if (code1 >= 0) {
	 Integer key = makeKey(code1, code2);
	 if (byCode.containsKey(key)) {
	    throw new KerningPairTableException(code1 + " " + code2,
	       "multiply defined kerning pair");
	 }
	 byCode.put(key, pair);
      } else {
	 resolveCode.add(pair);
      }
   }

   /**
    * Returns true, if a kerning pair with the given encoding values
    * is included in the table.
    * @param code1 encoding value of the first glyph.
    * @param code2 encoding value of the second glyph.
    */
   public boolean defined(int code1, int code2) {
      return byCode.containsKey(makeKey(code1, code2));
   }

   /**
    * Returns true, if a kerning pair with the given glyph names
    * is included in the table.
    * @param name1 name of the first glyph.
    * @param name2 name of the second glyph.
    */
   public boolean defined(String name1, String name2) {
      return byName.containsKey(makeKey(name1, name2));
   }

   /**
    * Return the kerning pair for the given encoding values.
    * Null is returned, if no such pair is in the table.
    * @param code1 encoding value of the first glyph.
    * @param code2 encoding value of the second glyph.
    */
   public KerningPair get(int code1, int code2) {
      return byCode.get(makeKey(code1, code2));
   }

   /**
    * Return the kerning pair for the given glyph names.
    * Null is returned, if no such pair is in the table.
    * @param name1 name of the first glyph.
    * @param name2 name of the second glyph.
    */
   public KerningPair get(String name1, String name2) {
      return byName.get(makeKey(name1, name2));
   }

   /**
    * Make kerning pairs accessible through names whose
    * pair has been specified by encoding values and vice
    * versa. Multiple invocations are permitted.
    * @param cms This character metrics table is used
    *            to perform the necessary mapping.
    */
   public void resolve(CharacterMetricsTable cms) {
      ListIterator<KerningPair> it;

      it = resolveName.listIterator(0); 
      while (it.hasNext()) {
	 KerningPair pair = it.next();
	 int code1 = pair.getCode1();
	 int code2 = pair.getCode2();
	 CharacterMetrics cm1 = cms.get(code1);
	 CharacterMetrics cm2 = cms.get(code2);
	 String name1 = cm1.getName();
	 String name2 = cm2.getName();
	 if (name1 != null && name2 != null) {
	    byName.put(makeKey(name1, name2), pair);
	    it.remove();
	 }
      }

      it = resolveCode.listIterator(0); 
      while (it.hasNext()) {
	 KerningPair pair = it.next();
	 String name1 = pair.getName1();
	 String name2 = pair.getName2();
	 CharacterMetrics cm1 = cms.get(name1);
	 CharacterMetrics cm2 = cms.get(name2);
	 int code1 = cm1.getCode();
	 int code2 = cm2.getCode();
	 if (code1 >= 0 && code2 >= 0) {
	    byCode.put(makeKey(code1, code2), pair);
	    it.remove();
	 }
      }
   }

   /**
    * Return a representation of the internal data structures
    * as printable string.
    */
   public String toString() {
      return
	 "kerning pairs by code values: " + byCode.toString() + "\n" +
	 "kerning pairs by names: " + byName.toString() + "\n";
   }
} // KerningPairTable
