package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * Implementation of the LineBreaker interface that breaks paragraphs
 * into lines using the best-fit algorithm. This algorithm processes
 * line by line like FirstFitLineBreaker but compares all acceptable
 * line break positions to take the one whose badness is minimal.
 */
public class BestFitLineBreaker extends BasicLineBreaker {

   private static double computeBadness(Width width, int parwidth) {
      double badness;

      if (width.getWidth() < parwidth) {
	 if (width.getStretchability() > 0) {
	    badness = ((double) parwidth - width.getWidth()) /
	       width.getStretchability();
	 } else {
	    // this is bad, but not as bad as shrinking over
	    // the the limits of shrinkability
	    badness = 100 + 2 * (parwidth - width.getWidth());
	 }
      } else {
	 if (width.getShrinkability() > 0) {
	    badness = ((double) width.getWidth() - parwidth) /
	       width.getShrinkability();
	    if (badness > 1) {
	       // do not shrink beyond the given shrinkability
	       badness = Item.INFINITY;
	    }
	 } else {
	    badness = Item.INFINITY / 2 + 2 * (width.getWidth() - parwidth);
	 }
      }
      return badness;
   }

   private static Width computeWidth(HorizontalSequence line,
	 Item bp) {
      Width width = line.getWidth();
      if (bp != null) {
	 Item hyphen = bp.getHyphen();
	 if (hyphen != null) {
	    width.add(hyphen);
	 }
      }
      return width;
   }

   /**
    * Apply the first-fit line-breaking algorithm to the given
    * horizontal sequence and return a vertical box with the
    * individual lines.
    * @param hseq The horizontal sequence of boxes that
    *             is to be split into lines.
    * @param parwidth The width of the paragraph in units (1/1000 points).
    * @param baselineskip The vertical distance between the baselines of
    *                     two subsequent lines in units (1/1000 points).
    */
   public Item breakParagraph(HorizontalSequence hseq,
	 int parwidth, int baselineskip) {
      VerticalBox vbox = new VerticalBox(baselineskip);

      HorizontalSequence line = null; // current line
      HorizontalSequence candidate = null;
      Item candidateHyphen = null;
      HorizontalSequence nextline = null; // beginning of next line
      double badness = 0; // of candidate
      Item bp = null; // last breakpoint
      for (HorizontalSequence word: hseq.getWords()) {
	 if (line == null) {
	    line = word;
	 } else {
	    line.add(bp); line.add(word);
	 }
	 if (candidate != null) {
	    if (nextline == null) {
	       nextline = word;
	    } else {
	       nextline.add(bp); nextline.add(word);
	    }
	 }

	 bp = word.getFollowingBreakpoint();
	 Width width = computeWidth(line, bp);
	 Item hyphen = bp.getHyphen();
	 double newbadness = computeBadness(width, parwidth);
	 if (candidate == null || newbadness < badness) {
	    candidate = line.clone();
	    candidateHyphen = hyphen;
	    nextline = null;
	    badness = newbadness;
	 }

	 if (width.getWidth() - width.getShrinkability() > parwidth) {
	    if (candidate == null) {
	       if (hyphen != null) {
		  line.add(hyphen);
	       }
	       add(vbox, line, parwidth);
	       line = null;
	    } else {
	       if (candidateHyphen != null) {
		  candidate.add(candidateHyphen);
	       }
	       add(vbox, candidate, parwidth);
	       line = nextline;
	       candidate = null; nextline = null;
	       candidateHyphen = null;
	    }
	 }
	 if (candidate == null && line != null) {
	    candidate = line.clone();
	    candidateHyphen = hyphen;
	    badness = computeBadness(computeWidth(candidate, bp), parwidth);
	    nextline = null;
	 }
      }
      if (line != null) {
	 line.add(bp);
	 add(vbox, line, parwidth);
      }
      return vbox;
   }
} // class BestFitLineBreaker
