package de.uniulm.mathematik.typo.items;

/**
 * Abstract base class for regular boxes, i.e. boxes that can neither
 * be shrunk nor stretched and which cannot be split. Extensions of
 * this class just need to implement the remaining abstract methods
 * getWidth(), getHeight(), getDepth(), and genPostScript().
 */
public abstract class Box implements Item {
   /**
    * Returns false.
    */
   final public boolean isPenalty() {
      return false;
   }
   /**
    * Returns false.
    */
   final public boolean isGlue() {
      return false;
   }
   /**
    * Returns true.
    */
   final public boolean isBox() {
      return true;
   }
   /**
    * Returns the width of the box in units (1/1000 points).
    */
   public abstract int getWidth();
   /**
    * Returns always 0, i.e. stretching is not supported.
    */
   public int getStretchability() {
      return 0;
   }
   /**
    * Returns always 0, i.e. shrinking is not supported.
    */
   public int getShrinkability() {
      return 0;
   }
   /**
    * Is a no-operation as shrinking is not supported.
    */
   public void shrink(int units) {}
   /**
    * Is a no-operation as stretching is not supported.
    */
   public void stretch(int units) {}
   /**
    * Returns the height of the box in units (1/1000 points).
    */
   public abstract int getHeight();
   /**
    * Returns the depth of the box in units (1/1000 points).
    */
   public abstract int getDepth();
   /**
    * Returns Item.INFINITY, i.e. this box cannot be split.
    */
   final public int getPenalty() {
      return Item.INFINITY;
   }
   /**
    * Returns false.
    */
   public boolean getPenaltyFlag() {
      return false;
   }
   /**
    * Returns null.
    */
   public Item getHyphen() {
      return null;
   }
   /**
    * Generate and return the string that represents this box
    * in PostScript text.
    */
   public abstract StringBuffer genPostScript(PostScriptContext context);
} // class Box
