package de.uniulm.mathematik.typo.items;

import java.util.*;

/**
 * Implementation of the LineBreaker interface that breaks paragraphs
 * into lines using the first-fit algorithm. This algorithm adds words
 * to the current line until it overflows. If it is possible, to squeeze
 * the line's content into the paragraph box width, it is done. Otherwise
 * the word that caused the overflow is moved to the beginning of the
 * following line. The current line is then stretched such that it
 * spans the intended paragraph box width.
 */
public class FirstFitLineBreaker extends BasicLineBreaker {
   /**
    * Apply the first-fit line-breaking algorithm to the given
    * horizontal sequence and return a vertical box with the
    * individual lines.
    * @param hseq The horizontal sequence of boxes that
    *             is to be split into lines.
    * @param parwidth The width of the paragraph in units (1/1000 points).
    * @param baselineskip The vertical distance between the baselines of
    *                     two subsequent lines in units (1/1000 points).
    */
   public Item breakParagraph(HorizontalSequence hseq,
	 int parwidth, int baselineskip) {
      VerticalBox vbox = new VerticalBox(baselineskip);

      HorizontalSequence line = null; // current line
      Item bp = null; // last breakpoint
      for (HorizontalSequence word: hseq.getWords()) {
	 if (line == null) {
	    line = word;
	 } else {
	    Width sum = new Width(line.getWidth());
	    sum.add(bp); sum.add(word.getWidth());
	    if (sum.getWidth() > parwidth) {
	       if (sum.getWidth() - sum.getShrinkability() <= parwidth) {
		  line.add(bp); line.add(word);
		  add(vbox, line, parwidth);
		  line = null;
	       } else {
		  Item hyphen = bp.getHyphen();
		  if (hyphen != null) {
		     line.add(hyphen);
		  }
		  add(vbox, line, parwidth);
		  line = word;
	       }
	    } else {
	       line.add(bp); line.add(word);
	    }
	 }
	 bp = word.getFollowingBreakpoint();
      }
      if (line != null) {
	 line.add(bp);
	 add(vbox, line, parwidth);
      }
      return vbox;
   }
} // class FirstFitLineBreaker
