package de.uniulm.mathematik.typo.items;

import de.uniulm.mathematik.typo.items.*;

/**
 * Special item class that implements glue. Glue has no visible
 * contents but serves as breakpoint and provides flexibility as
 * it is stretchable and shrinkable.
 */
public class Glue implements Item {
   private int originalWidth;
   private int width;
   private int stretchability;
   private int shrinkability;

   private int intoRange(int value) {
      if (value > Item.INFINITY) {
	 return Item.INFINITY;
      } else if (value < - Item.INFINITY) {
	 return -Item.INFINITY;
      } else {
	 return value;
      }
   }

   /**
    * Construct a glue object with the given parameters:
    * @param width preferred width of the glue object in units (1/1000 point).
    * @param stretchability non-negative amount by which the width of this
    *                       object can be extended without causing
    *                       typographically poor output.
    * @param shrinkability non-negative amount by which the width of this
    *                      object can be shrunk without running into
    *                      typographic problems.
    */
   public Glue(int width, int stretchability, int shrinkability) {
      this.width = intoRange(width);
      originalWidth = this.width;
      assert(stretchability >= 0);
      this.stretchability = intoRange(stretchability);
      assert(shrinkability >= 0);
      this.shrinkability = intoRange(shrinkability);
   }

   /**
    * Returns true.
    */
   final public boolean isGlue() {
      return true;
   }

   /**
    * Returns false.
    */
   final public boolean isPenalty() {
      return false;
   }

   /**
    * Returns false.
    */
   final public boolean isBox() {
      return false;
   }

   /**
    * Returns the current width in units (1/1000 points). This width
    * is computed from the original width plus any stretching amounts
    * and minus any shrinking amounts. Negative widths are possible.
    */
   public int getWidth() {
      return width;
   }

   /**
    * Return the largest acceptable stretching amount in units.
    */
   public int getStretchability() {
      return stretchability;
   }

   /**
    * Returns the largest acceptable shrinking amount in units.
    */
   public int getShrinkability() {
      return shrinkability;
   }

   /**
    * Stretch the width of this glue object by the given amount in units.
    * Multiple invocations cause this object to be stretched multiple times.
    * There is no safeguard that prevents to specify an amount which
    * exceeds that returned by getStretchability(). The amount must
    * be non-negative.
    */
   public void stretch(int units) {
      assert(units >= 0);
      width = intoRange(originalWidth + units);
      if (units > stretchability) {
	 shrinkability += stretchability; stretchability = 0;
      } else {
	 stretchability -= units; shrinkability += units;
      }
   }

   /**
    * Shrinks the width of this glue object by the given amount in units.
    * Multiple invocations cause this object to be shrunk multiple times.
    * There is no safeguard that prevents to specify an amount which
    * exceeds that returned by getShrinkability(). The amount must
    * be non-negative.
    */
   public void shrink(int units) {
      assert(units >= 0);
      width = intoRange(originalWidth - units);
      if (units > shrinkability) {
	 stretchability += shrinkability; shrinkability = 0;
      } else {
	 shrinkability -= units; stretchability += units;
      }
   }

   /**
    * Returns a height of 0.
    */
   public int getHeight() {
      return 0;
   }

   /**
    * Returns a depth of 0.
    */
   public int getDepth() {
      return 0;
   }

   /**
    * The penalty value for splitting this object is 0,
    * i.e. the text can be happily split at this location but
    * there is special urge to do this.
    */
   public int getPenalty() {
      return 0;
   }

   /**
    * Returns false, i.e. there is no problem with consecutive
    * splittings of this kind.
    */
   final public boolean getPenaltyFlag() {
      return false;
   }

   /**
    * Returns null.
    */
   public Item getHyphen() {
      return null;
   }

   /**
    * Generates a PostScript text that moves relatively along
    * the horizontal baseline according to the current width.
    */
   public StringBuffer genPostScript(PostScriptContext context) {
      StringBuffer result = context.closeString();
      if (width != 0) {
	 result.append((double) width / 1000);
	 result.append(" 0 rmoveto\n");
      }
      return result;
   }

   /**
    * Return a printable summary of the glue object.
    */
   public String toString() {
      StringBuffer result = new StringBuffer("glue box {");
      result.append("width = " + width);
      result.append(", stretchability = " + stretchability);
      result.append(", shrinkability = " + shrinkability);
      result.append("}");
      return result.toString();
   }
} // class Glue
