package de.uniulm.mathematik.typo.items;

import java.lang.*;

/**
 * Interface for all items. Items fall into one of three categories:
 * <UL>
 * <LI>Regular boxes, e.g. items representing individual glyphs
 *     or fixed spacings (like kerning).
 * <LI>Glue items, i.e. items that represent no visible output
 *     but stretchable or shrinkable horizontal spaces.
 * <LI>Penalty items, i.e. items that indicate a possible position
 *     where a sequence can be split into lines.
 * </UL>
 * <P>
 * All items represent a rectangle in 2-dimensional space. The
 * width specifies the extent along the x-axis, the height and
 * the depth the extents along the y-axis:
 *
 * <PRE>
 *       y ^
 *         |
 *         |
 *  height +------------------+
 *         |                  |
 *         |                  |
 *         |                  |
 *         |                  |
 *         |                  |
 *         |                  |
 *         |                  |
 *     ----+------------------+---------------&gt; x
 *         |                  |
 *         |                  |
 *   depth +------------------+
 *                          width
 * </PRE>
 * <P>Please note that the width can be negative but the height and
 * the depth are assumed to be non-negative.
 *
 * When an item is to be drawn using the genPostScript() method, the
 * pencil is assumed to be at the origin. When finished, the pencil
 * should be left at (0, width). All movements are to be done relatively
 * in PostScript.
 */
public interface Item {
   /**
    * All measures are confined into the interval [-INFINITY, INFINITY].
    */
   public final int INFINITY = Integer.MAX_VALUE / 2;
   /**
    * Returns true, if this item is a penalty item.
    */
   public boolean isPenalty();
   /**
    * Returns true, if this item is glue item.
    */
   public boolean isGlue();
   /**
    * Returns true, if this item is a regular box.
    */
   public boolean isBox();
   /**
    * Return the width of this item in units (1/1000 points).
    */
   public int getWidth();
   /**
    * Return the stretchability of this item in units (1/1000 points).
    */
   public int getStretchability();
   /**
    * Return the shrinkability of this item in units (1/1000 points).
    */
   public int getShrinkability();
   /**
    * Shrink this item by the given amount of units. This operation
    * is supported by glue items only.
    */
   public void shrink(int units);
   /**
    * Stretch this item by the given amount of units. This operation
    * is supported by glue items only.
    */
   public void stretch(int units);
   /**
    * Return the height of this item in units (1/1000 points).
    */
   public int getHeight();
   /**
    * Return the depth of this item in units (1/1000 points).
    */
   public int getDepth();
   /**
    * Return the penalty value, if it is a penalty box. Otherwise INFINITY.
    * The higher the penalty value, the uglier a break at this
    * point is expected to be. A penalty value of -INFINITY means that
    * a break at that point is mandatory.
    */
   public int getPenalty();
   /**
    * Consecutive line breaks at flagged penalty items are considered
    * ugly.
    */
   public boolean getPenaltyFlag();
   /**
    * Return an item that is to be inserted at the end of
    * the upper line if a line is to be split at this point.
    * Null may be returned if there is nothing to be inserted
    * or if this is not a penalty.
    */
   public Item getHyphen();
   /**
    * Return the PostScript code to draw this box.
    */
   public StringBuffer genPostScript(PostScriptContext context);
} // class Item
