package de.uniulm.mathematik.typo.items;

import java.lang.*;
import java.util.*;

/**
 * This is the standard implementation of a horizontal sequence.
 */
public class SimpleHorizontalSequence implements HorizontalSequence {

   private class Node {
      public Item item;
      public Node prev, next;

      public boolean isBreakpoint;
      // in case of breakpoints only:
      public Node begin; // first node behind the last breakpoint
      public Node nextbp; // points to the next breakpoint
      public Width wwidth; // width of the preceding word

      public boolean isGlue;
      // in case of glue items only:
      public Node nextglue; // points to the next glue item

      public Node(Item item) {
	 this.item = item;
	 next = null; prev = null;
	 wwidth = null;
	 isBreakpoint = false;
      }
   }

   private abstract class BasicItemIterator implements Iterator<Item> {
      protected Node node;
      protected Node tail;
      protected Iterator<Item> extension;
      protected boolean inExtension;

      public boolean hasNext() {
	 if (inExtension) {
	    return extension.hasNext();
	 } else {
	    return node != null;
	 }
      }

      public abstract void nextInList();

      public Item next() throws NoSuchElementException {
	 if (inExtension) {
	    return extension.next();
	 } else {
	    if (node == null) {
	       throw new NoSuchElementException("end of horizontal sequence");
	    } else {
	       Item item = node.item;
	       if (node == tail) {
		  node = null;
	       } else {
		  nextInList();
	       }
	       if (node == null && extension != null) {
		  inExtension = true;
	       }
	       return item;
	    }
	 }
      }

      public void remove() throws UnsupportedOperationException {
	 throw new UnsupportedOperationException();
      }
   }

   private class ItemIterator extends BasicItemIterator {
      public ItemIterator(Node head) {
	 node = head; tail = null; extension = null; inExtension = false;
      }

      public ItemIterator(Node head, Node tail) {
	 this(head);
	 this.tail = tail;
      }

      public ItemIterator(Node head, Node tail, Iterator<Item> extension) {
	 this(head, tail);
	 this.extension = extension;
	 if (head == null) {
	    inExtension = true;
	 }
      }

      public void nextInList() {
	 node = node.next;
      }
   }

   private class GlueItemIterator extends BasicItemIterator {
      public GlueItemIterator(Node head) {
	 node = head; tail = null; extension = null; inExtension = false;
      }

      public GlueItemIterator(Node head, Node tail) {
	 this(head);
	 this.tail = tail;
      }

      public GlueItemIterator(Node head, Node tail, Iterator<Item> extension) {
	 this(head, tail);
	 this.extension = extension;
      }

      public void nextInList() {
	 node = node.nextglue;
      }
   }

   private class WordIterator implements Iterator<HorizontalSequence> {
      private Node bp;
      private Node lastbp;
      private Iterator<HorizontalSequence> extension;
      private boolean inExtension;

      public WordIterator(Node bp) {
	 this.bp = bp; lastbp = null; extension = null; inExtension = false;
      }

      public WordIterator(Node bp, Node lastbp) {
	 this.bp = bp; this.lastbp = lastbp; extension = null;
	 inExtension = false;
      }

      public WordIterator(Node bp, Node lastbp,
	    Iterator<HorizontalSequence> extension) {
	 assert(bp != null);
	 this.bp = bp; this.lastbp = lastbp; this.extension = extension;
	 inExtension = false;
	 lastglue = null; firstglue = null;
      }

      public boolean hasNext() {
	 if (inExtension) {
	    return extension.hasNext();
	 } else {
	    return bp != null;
	 }
      }

      public HorizontalSequence next() throws NoSuchElementException {
	 if (inExtension) {
	    return extension.next();
	 } else {
	    if (bp == null) {
	       throw new NoSuchElementException();
	    } else {
	       HorizontalSequence subseq = new Subsequence(bp);
	       if (bp == lastbp) {
		  bp = null;
	       } else {
		  bp = bp.nextbp;
	       }
	       if (bp == null && extension != null) {
		  inExtension = true;
	       }
	       return subseq;
	    }
	 }
      }

      public void remove() throws UnsupportedOperationException {
	 throw new UnsupportedOperationException();
      }
   }

   private class Subsequence implements HorizontalSequence {
      private Node bp;
      private Node lastbp;
      private Node head, tail;
      private Width width;
      private boolean strictSubrange;
      private boolean extended;
      private HorizontalSequence extension;
      private Node lastglue;
      private Node firstglue;

      public Subsequence(Node bp) {
	 assert(bp.isBreakpoint);
	 this.bp = bp; lastbp = bp;
	 head = bp.begin;
	 tail = bp.prev;
	 width = new Width(bp.wwidth);
	 extended = false;
	 extension = null;
	 strictSubrange = true;
	 lastglue = null;
	 firstglue = null;
      }

      public Subsequence(Node firstbp, Node lastbp, Width width,
	    Node firstglue, Node lastglue) {
	 assert(firstbp.isBreakpoint);
	 assert(lastbp.isBreakpoint);
	 this.bp = bp; this.lastbp = lastbp;
	 head = bp.begin; tail = lastbp;
	 this.width = width;
	 extended = true;
	 extension = null;
	 strictSubrange = true;
	 this.firstglue = firstglue;
	 this.lastglue = lastglue;
      }

      private Subsequence(Node firstbp, Node lastbp, Node head, Node tail,
	    Width width, boolean strictSubrange, boolean extended,
	    HorizontalSequence extension, Node firstglue, Node lastglue) {
	 this.bp = firstbp; this.lastbp = lastbp;
	 this.head = head; this.tail = tail; this.width = new Width(width);
	 this.strictSubrange = strictSubrange; this.extended = extended;
	 if (extension == null) {
	    this.extension = null;
	 } else {
	    this.extension = extension.clone();
	 }
	 this.firstglue = firstglue; this.lastglue = lastglue;
      }

      public void add(Item item) {
	 width.add(item);
	 if (strictSubrange && head == null && item == bp.item) {
	    head = bp; tail = bp; lastbp = bp;
	 } else if (strictSubrange &&
	       tail != null && tail.next != null && tail.next.item == item) {
	    tail = tail.next;
	    if (tail.isBreakpoint) {
	       lastbp = tail;
	    }
	 } else {
	    strictSubrange = false;
	    if (extension == null) {
	       extension = new SimpleHorizontalSequence();
	       if (!tail.isBreakpoint) {
		  // take care that words remain recognizable
		  if (tail.next == lastbp) {
		     // copy the entire sequence to the extension
		     for (Node p = head; p != tail.next; p = p.next) {
			extension.add(p.item);
		     }
		     // cut off the original subsequence
		     head = null; tail = null; lastbp = null; bp = null;
		  } else {
		     // move items since last breakpoint to
		     // the extension sequence
		     Node p = lastbp.next;
		     for(;;) {
			extension.add(p.item);
			if (p == tail) {
			   break;
			}
			p = p.next;
		     }
		     tail = lastbp;
		  }
	       }
	    }
	    extension.add(item);
	 }
	 if (strictSubrange && tail.isGlue) {
	    if (firstglue == null) {
	       firstglue = tail;
	    }
	    lastglue = tail;
	 }
	 extended = true;
      }

      public void add(HorizontalSequence hseq) {
	 for (Item item: hseq) {
	    add(item);
	 }
      }

      public Width getWidth() {
	 return new Width(width);
      }

      public IterableIterator<HorizontalSequence> getWords() {
	 if (extension != null) {
	    if (head == null) {
	       return extension.getWords();
	    } else {
	       return new IterableIterator<HorizontalSequence>
			(new WordIterator(bp, lastbp, extension.getWords()));
	    }
	 } else {
	    return new IterableIterator<HorizontalSequence>
			   (new WordIterator(bp, lastbp));
	 }
      }

      public Item getFollowingBreakpoint() {
	 if (extended) {
	    return null;
	 } else {
	    return bp.item;
	 }
      }

      public IterableIterator<Item> getGlueItems() {
	 if (!extended) {
	    return new IterableIterator<Item>(new GlueItemIterator(null));
	 } else if (firstglue == null && extension != null) {
	    return extension.getGlueItems();
	 } else if (extension == null) {
	    return new IterableIterator<Item>(new GlueItemIterator(firstglue,
	       lastglue));
	 } else {
	    return new IterableIterator<Item>(new GlueItemIterator(firstglue,
	       lastglue, extension.getGlueItems()));
	 }
      }

      public HorizontalSequence clone() {
	 return new Subsequence(bp, lastbp, head, tail,
			width, strictSubrange, extended,
			extension, firstglue, lastglue);
      }

      public Iterator<Item> iterator() {
	 if (head == null && extension != null) {
	    return extension.iterator();
	 } else if (extension == null) {
	    return new ItemIterator(head, tail);
	 } else {
	    return new ItemIterator(head, tail, extension.iterator());
	 }
      }
   }

   private Node head, tail; // linear list of boxes
   private boolean beginWord;
   private Node begin;
   private Node lastbp;
   private Width width;
   private Width wwidth; // width of current word
   private Node firstbp;
   private Node lastglue;
   private Node firstglue;

   public SimpleHorizontalSequence() {
      head = null; tail = null;
      begin = null; beginWord = true;
      lastbp = null; firstbp = null;
      lastglue = null; firstglue = null;
      width = new Width();
      wwidth = new Width();
   }

   /**
    * Legal breakpoints are
    * <UL>
    * <LI>penalties with a penalty value less than INFINITY and
    * <LI>glue boxes that follow a regular box
    * </UL>
    */
   protected boolean isBreakpoint(Node node) {
      return node.prev != null &&
             (node.item.isPenalty() &&
		node.item.getPenalty() < Item.INFINITY ||
	     node.item.isGlue() && node.prev.item.isBox());
   }

   /**
    * Append the given item at the end of the horizontal sequence.
    * @param item to be added item.
    */
   public void add(Item item) {
      Node node = new Node(item);
      if (head == null) {
	 head = node;
      } else {
	 tail.next = node;
      }
      node.prev = tail;
      tail = node;
      width.add(item);
      node.begin = begin;

      if (item.isGlue()) {
	 if (firstglue == null) {
	    firstglue = node;
	 }
	 if (lastglue != null) {
	    lastglue.nextglue = node;
	 }
	 lastglue = node;
	 node.nextglue = null;
	 node.isGlue = true;
      }
      if (isBreakpoint(node)) {
	 node.isBreakpoint = true;
	 if (lastbp != null) {
	    lastbp.nextbp = node;
	 }
	 if (firstbp == null) {
	    firstbp = node;
	 }
	 node.wwidth = wwidth;
	 wwidth = new Width();
	 lastbp = node;
	 beginWord = true;
	 begin = null;
      } else {
	 wwidth.add(item);
	 if (beginWord) {
	    begin = node;
	    beginWord = false;
	 }
      }
   }

   /**
    * Append the given items out of the given horizontal sequence to
    * the end.
    * @param hseq to be added sequence.
    */
   public void add(HorizontalSequence hseq) {
      for (Item item: hseq) {
	 add(item);
      }
   }

   /**
    * Return the overall width of the sequence.
    */
   public Width getWidth() {
      Width clonedWidth = new Width(width);
      return clonedWidth;
   }

   public IterableIterator<HorizontalSequence> getWords() {
      return new IterableIterator<HorizontalSequence>
		     (new WordIterator(firstbp));
   }

   /**
    * Return the breakpoint following a subsequence. This returns
    * null if there is no such breakpoint following.
    */
   public Item getFollowingBreakpoint() {
      return null;
   }

   /**
    * Return glue items, i.e. items that can be shrunk or extended.
    */
   public IterableIterator<Item> getGlueItems() {
      return new IterableIterator<Item>(new GlueItemIterator(firstglue));
   }

   /**
    * Return iterator over this horizontal sequence.
    */
   public Iterator<Item> iterator() {
      return new ItemIterator(head);
   }

   /**
    * Return an independent cloned copy of this horizontal sequence.
    * Please note that just the collection is cloned but not the
    * individual items.
    */
   public HorizontalSequence clone() {
      if (head == null) {
	 return new SimpleHorizontalSequence();
      } else if (tail.isBreakpoint) {
	 return new Subsequence(firstbp, lastbp, new Width(width),
	    firstglue, lastglue);
      } else {
	 // inefficient variant
	 HorizontalSequence result = new SimpleHorizontalSequence();
	 for (Item item: this) {
	    result.add(item);
	 }
	 return result;
      }
   }
} // class HorizontalSequence
